import gc
import psutil
from .actor import Actor
from .. import sleep


class GarbageCollector(Actor):
    '''Disabling Python's automatic garbage collection to optimize performance.'''
    interval: float = 10 * 60
    mbyte_limit: float = 300

    starting_msg = 'performing garbage collection'
    finished_msg = 'finished garbage collection'

    def __init__(self) -> None:
        super().__init__()
        gc.disable()

    async def step(self):
        await super().step()
        if psutil.virtual_memory().free < self.mbyte_limit * 1000000:
            self.log.warning(f'less than {self.mbyte_limit} mb of memory remaining -> {self.starting_msg}')
            gc.collect()
            await sleep(1)  # NOTE yield execution to make sure all warnings appear before we send the "finish" message
            self.log.warning(self.finished_msg)



from typing import Optional
from .. import sleep
from ..communication import Communication
from ..world import Velocity, World
from . import CommunicatingHardware


class RobotBrain(CommunicatingHardware):

    def __init__(self, world: World, communication: Communication):
        super().__init__(world, communication)
        self.waiting_list: dict[str, Optional[str]] = {}

    async def configure(self, filepath: str = 'lizard.txt'):
        await super().configure()
        with open(filepath) as f:
            expander = False
            await self.send(f'!-')
            for line in f.read().splitlines():
                if line == '---':
                    expander = True
                    await self.send('!>!-')
                else:
                    if expander:
                        await self.send(f'!>!+{line}')
                    else:
                        await self.send(f'!+{line}')
            if expander:
                await self.send(f'!>!.')
                await self.send(f'!>core.restart()')
            await self.send(f'!.')
            await self.restart()

    async def restart(self):
        await super().restart()
        await self.send(f'core.restart()')

    async def drive(self, linear: float, angular: float):
        await super().drive(linear, angular)
        await self.send(f'wheels.speed({linear}, {angular})')

    async def stop(self):
        await super().stop()
        await self.send('wheels.off()')

    async def update(self):
        await super().update()
        millis = None
        while True:
            line = self.check(await self.communication.read())
            if line is None or not line:
                break
            words = line.split()
            if not words:
                continue
            first = words.pop(0)
            if first in self.waiting_list:
                self.waiting_list[first] = line
                continue
            if first == 'core':
                millis = float(words.pop(0))
                if self.world.robot.clock_offset is None:
                    continue
                self.world.robot.hardware_time = millis / 1000 + self.world.robot.clock_offset
                self.parse_core(words)
            self.parse_line(line)
        if millis is not None:
            self.world.robot.clock_offset = self.world.time - millis / 1000

    def parse_core(self, words: list[str]):
        self.world.robot.odometry.append(Velocity(
            linear=float(words.pop(0)),
            angular=float(words.pop(0)),
            time=self.world.robot.hardware_time,
        ))

    def parse_line(self, line: str):
        pass

    async def send(self, msg: str):
        await self.communication.send_async(self.augment(msg))

    async def send_and_await(self, msg: str, ack: str, *, timeout: float = float('inf')) -> Optional[str]:
        self.waiting_list[ack] = None
        await self.send(msg)
        t0 = self.world.time
        while self.waiting_list.get(ack) is None and self.world.time < t0 + timeout:
            await sleep(0.1)
        return self.waiting_list.pop(ack) if ack in self.waiting_list else None

    @staticmethod
    def augment(line: str) -> str:
        checksum = 0
        for c in line:
            checksum ^= ord(c)
        return f'{line}@{checksum:02x}'

    @staticmethod
    def check(line: Optional[str]) -> str:
        if line is None:
            return ""
        if line[-3:-2] == '@':
            check = int(line[-2:], 16)
            line = line[:-3]
            checksum = 0
            for c in line:
                checksum ^= ord(c)
            if checksum != check:
                return None
        return line



import asyncio
from typing import Callable
import subprocess
from concurrent.futures import ProcessPoolExecutor
import logging
import uuid


process_pool = ProcessPoolExecutor()
running_processes = []  # NOTE is used in rosys.test.Runtime to advance time slower until computation is done
log = logging.getLogger('rosys.run')


async def io_bound(callback: Callable, *args: any):
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, callback, *args)


async def cpu_bound(callback: Callable, *args: any):
    id = str(uuid.uuid4())
    running_processes.append(id)
    try:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(process_pool, callback, *args)
    except RuntimeError as e:
        if 'cannot schedule new futures after shutdown' not in str(e):
            raise
    finally:
        running_processes.remove(id)


async def sh(command: list[str]) -> str:
    '''executes a shell command
    command: a sequence of program arguments as subprocess.Popen requires
    returns: stdout
    '''
    def run() -> str:
        proc = subprocess.Popen(command, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.STDOUT)
        stdout, *_ = proc.communicate()
        return stdout.decode()
    #self.log.debug('executing sh command: ' + ' '.join(command))
    return await io_bound(run)



from pydantic import BaseModel
from typing import Any, Optional
import numpy as np
import cv2
from .image import ImageSize
from .point import Point
from .point3d import Point3d
from .rotation import Rotation


class Intrinsics(BaseModel):
    matrix: list[list[float]]
    distortion: list[float]
    rotation: Rotation
    size: ImageSize


class Extrinsics(BaseModel):
    tilt: Optional[Rotation]
    yaw: float = 0.0
    translation: list[float] = [0.0, 0.0, 1.0]


class Calibration(BaseModel):
    intrinsics: Intrinsics
    extrinsics: Extrinsics = Extrinsics()

    @property
    def rotation(self) -> Rotation:
        tilt = self.extrinsics.tilt or Rotation.zero()
        return Rotation.from_euler(0, 0, self.extrinsics.yaw) * tilt * self.intrinsics.rotation

    @property
    def rotation_array(self) -> Any:
        tilt = self.extrinsics.tilt or Rotation.zero()
        yaw = self.extrinsics.yaw
        return np.dot([[np.cos(yaw), -np.sin(yaw), 0], [np.sin(yaw), np.cos(yaw), 0], [0, 0, 1]], tilt.R) @ self.intrinsics.rotation.R

    @property
    def is_complete(self) -> bool:
        return self.extrinsics.tilt is not None

    def project_to_image(self, world_point: Point3d) -> Point:
        world_points = np.array([world_point.tuple])
        R = self.rotation_array
        Rod = cv2.Rodrigues(R.T)[0]
        t = -R.T @ self.extrinsics.translation
        K = np.array(self.intrinsics.matrix)
        D = np.array(self.intrinsics.distortion)
        image_points, _ = cv2.projectPoints(world_points, Rod, t, K, D)
        return Point(x=image_points[0, 0, 0], y=image_points[0, 0, 1])

    def project_array_to_image(self, world_points: Any) -> Any:
        R = self.rotation_array
        Rod = cv2.Rodrigues(R.T)[0]
        t = -R.T @ self.extrinsics.translation
        K = np.array(self.intrinsics.matrix)
        D = np.array(self.intrinsics.distortion)
        image_points, _ = cv2.projectPoints(world_points, Rod, t, K, D)
        return image_points.reshape(-1, 2)

    def project_from_image(self, image_point: Point, target_height: float = 0) -> Optional[Point3d]:
        K = np.array(self.intrinsics.matrix)
        D = np.array(self.intrinsics.distortion)
        image_points = np.array(image_point.tuple, dtype=np.float32).reshape(1, 1, 2)
        image_points_ = cv2.undistortPoints(image_points, K, D).reshape(-1, 2)
        image_points__ = cv2.convertPointsToHomogeneous(image_points_).reshape(-1, 3)
        objPoints = image_points__ @ self.rotation_array.T
        Z = self.extrinsics.translation[-1]
        t = np.array(self.extrinsics.translation)
        floorPoints = t.T - objPoints * Z / objPoints[:, 2:]

        reprojection = self.project_to_image(Point3d(x=floorPoints[0, 0], y=floorPoints[0, 1], z=0))
        if objPoints[0, -1] * np.sign(Z) > 0 or reprojection.distance(image_point) > 2:
            return None

        X, Y, _ = floorPoints[0]
        X_ = (X - t[0]) / (t[2] + target_height) * t[2] + t[0]
        Y_ = (Y - t[1]) / (t[2] + target_height) * t[2] + t[1]
        return Point3d(x=X_, y=Y_, z=target_height)

    @staticmethod
    def from_points(world_points: list[Point3d], image_points: list[Point], image_size: ImageSize, f0: float):
        world_points = np.array([p.tuple for p in world_points], dtype=np.float32).reshape(1, -1, 3)
        image_points = np.array([p.tuple for p in image_points], dtype=np.float32).reshape(1, -1, 2)
        K0 = np.array([[f0, 0, image_size.width / 2], [0, f0, image_size.height / 2], [0, 0, 1]], dtype=np.float32)
        flags = cv2.CALIB_USE_INTRINSIC_GUESS
        _, K, D, rvecs, tvecs = cv2.calibrateCamera(world_points, image_points, image_size.tuple, K0, None, flags=flags)

        rotation0 = Rotation(R=np.eye(3).tolist())
        intrinsics = Intrinsics(matrix=K.tolist(), distortion=D.tolist()[0], rotation=rotation0, size=image_size)

        rotation = Rotation.from_rvec(rvecs[0]).T
        translation = (-np.array(rotation.R).dot(tvecs)).flatten().tolist()
        extrinsics = Extrinsics(tilt=rotation, translation=translation)

        return Calibration(intrinsics=intrinsics, extrinsics=extrinsics)



from typing import Optional
from pydantic import Field
from .image import ImageSize
from .camera import Camera


class UsbCamera(Camera):
    capture: bool = True
    detect: bool = False
    connected: bool = False
    resolution: Optional[ImageSize] = Field(None, exclude=True)



import numpy as np
from .. import sleep
from ..hardware import Hardware
from ..world import World


async def drive_arc(world: World, hardware: Hardware):
    while world.robot.prediction.x < 2:
        await hardware.drive(1, np.deg2rad(25))
        await sleep(0.1)

    await hardware.stop()



from typing import Optional
import socketio
from collections import deque
from .communication import Communication


class WebCommunication(Communication):
    '''Remote connection to the Robot Brains Esp.

    This makes it possible to keep developing on your fast computer
    while communicating with the hardware components connected to a physical Robot Brain.
    '''

    ip: str = '192.168.43.1'
    port: int = 8081
    host: str = f'ws://{ip}:{port}'

    def __init__(self):
        super().__init__()
        self.buffer: deque[str] = deque()
        self.sio = socketio.AsyncClient()

        @self.sio.event
        def connect():
            assert self.sio.transport() == 'websocket'

        @self.sio.event
        def read(msg):
            self.buffer.append(msg)

    @classmethod
    def is_possible(cls) -> bool:
        try:
            sio = socketio.Client(reconnection=False)
            sio.on('connect', sio.disconnect)
            sio.connect(cls.host)
            return True
        except:
            return False

    async def read(self) -> Optional[str]:
        if not self.sio.connected:
            await self.sio.connect(self.host)
        if self.buffer:
            return self.buffer.pop()

    async def send_async(self, line: str):
        await self.sio.emit('write', line)

    async def tear_down(self):
        await self.sio.disconnect()



from .actor import Actor
from ..world import Image, UsbCamera
from .. import event


class UsbCameraSimulator(Actor):
    interval: float = 1

    async def step(self):
        await super().step()
        while len(self.world.usb_cameras) < 2:
            uid = f'simulated_cam_{len(self.world.usb_cameras)}'
            self.world.usb_cameras[uid] = UsbCamera(id=uid)
            await event.call(event.Id.NEW_CAMERA, self.world.usb_cameras[uid])

        for camera in self.world.usb_cameras.values():
            camera.images.append(Image.create_placeholder(f'{camera.id}: {self.world.time:.2f}', time=self.world.time))



import asyncio
from typing import Coroutine, Optional
from .. import event, task_logger
from ..automations import Automation
from . import Actor


class Automator(Actor):
    def __init__(self) -> None:
        super().__init__()
        self.automation: Optional[Automation] = None
        event.register(event.Id.PAUSE_AUTOMATION, self.pause)

    @property
    def is_stopped(self) -> bool:
        return self.automation is None or self.automation.is_stopped

    @property
    def is_running(self) -> bool:
        return self.automation is not None and self.automation.is_running

    @property
    def is_paused(self) -> bool:
        return self.automation is not None and self.automation.is_paused

    def start(self, coro: Coroutine):
        self.stop(because='new automation starts')
        self.automation = Automation(coro, self._handle_exception, on_complete=self._on_complete)
        task_logger.create_task(asyncio.wait([self.automation]), name='automation')

    def pause(self, because: str):
        if self.is_running:
            self.automation.pause()
            event.emit(event.Id.AUTOMATION_PAUSED, because)
            event.emit(event.Id.NEW_NOTIFICATION, f'automation paused because {because}')

    def resume(self):
        if self.is_paused:
            self.automation.resume()

    def stop(self, because: str):
        if not self.is_stopped:
            self.automation.stop()
            event.emit(event.Id.AUTOMATION_PAUSED, because)
            event.emit(event.Id.NEW_NOTIFICATION, f'automation stopped because {because}')

    def _handle_exception(self, e: Exception):
        self.stop(because='an exception occurred in an automation')

    def _on_complete(self):
        event.emit(event.Id.NEW_NOTIFICATION, f'automation completed')

    async def tear_down(self):
        await super().tear_down()
        self.stop(because='automator is shutting down')



from __future__ import annotations
from pydantic import BaseModel
from typing import Optional
import PIL.Image
import PIL.ImageDraw
import io
import urllib.parse
from .detection import Detection


class ImageSize(BaseModel):
    width: int
    height: int

    @property
    def tuple(self):
        return (self.width, self.height)


class Image(BaseModel):
    camera_id: str
    size: ImageSize
    time: float  # World time of recording
    data: Optional[bytes] = None
    detections: Optional[list[Detection]] = None

    @property
    def url(self) -> str:
        return f'camera/{urllib.parse.quote_plus(self.camera_id)}/{self.time}'

    @property
    def id(self) -> str:
        return f'{self.camera_id}/{self.time}'

    @staticmethod
    def create_placeholder(text: str, time: float = None) -> Image:
        img = PIL.Image.new('RGB', (260, 200), color=(73, 109, 137))
        d = PIL.ImageDraw.Draw(img)
        d.text((img.width/2-len(text)*3, img.height/2-5), text, fill=(255, 255, 255))
        bytesio = io.BytesIO()
        img.save(bytesio, format='PNG')
        image = Image(camera_id='no_cam_id', time=time or 0, size=ImageSize(width=img.width, height=img.height))
        image.data = bytesio.getvalue()
        return image

    def __hash__(self) -> int:
        return hash(self.id)



from nicegui.ui import Ui
import logging
import html
import numpy as np
from ..actors import AsyncioMonitor
from .. import run

log = logging.getLogger('rosys.asyncio')


def prepare_data(timings):
    def calc_box(data: list):
        return {
            'low': min(data),
            'q1': np.percentile(data, 25),
            'median': np.percentile(data, 50),
            'q3': np.percentile(data, 75),
            'high': max(data),
        }
    timings = dict(sorted(timings.items(), key=lambda i: len(i[1]), reverse=True))
    names = [f'{html.escape(n)} ({len(w)} warnings):<br>{html.escape(w[0].details)}' for n, w in timings.items()]
    return names, [calc_box([w.duration * 1000 for w in t]) for t in timings.values()]


class AsyncioPage:
    ui: Ui = None
    asyncio_monitor: AsyncioMonitor

    def __init__(self) -> None:
        with self.ui.page('/asyncio'):
            with self.ui.row():
                self.ui.button('load', on_click=self.update)
                self.info_label = self.ui.label()
            self.chart = self.ui.chart(options={
                'title': False,
                'chart': {'type': 'boxplot'},
                'xAxis': {'categories': []},
                'yAxis': {'title': {'text': 'ms on UI thread'}},
                'series': {'data': []},
                'navigation': {'buttonOptions': {'enabled': False}},
                'legend': False,
                'credits': False,
            }).classes('fit').style('height:400px')

    async def update(self):
        names, data = await run.cpu_bound(prepare_data, self.asyncio_monitor.timings)
        self.chart.options['xAxis']['categories'][:] = names
        self.chart.options['series']['data'][:] = data
        await self.chart.view.update()



from nicegui.ui import Ui
from .. import event
from ..communication import Communication, SerialCommunication


class LizardSerialDebug:
    ui: Ui = None  # will be set by rosys.ui.configure
    communication: Communication = None  # will be set by rosys.ui.configure

    def __init__(self) -> None:
        if not isinstance(self.communication, SerialCommunication):
            return
        self.ui.switch('Lizard', value=self.communication.serial.isOpen(), on_change=self.toggle_communication)
        self.input = self.ui.input(on_change=self.submit_message)

    async def submit_message(self):
        await self.communication.send_async(self.input.value)
        self.input.value = ''
        await self.input.view.update()

    async def toggle_communication(self, status):
        assert isinstance(self.communication, SerialCommunication)
        if status.value:
            self.communication.connect()
            await event.call(event.Id.NEW_NOTIFICATION, 'connected to Lizard')
        else:
            await event.call(event.Id.PAUSE_AUTOMATION, 'communication is deactivated')
            self.communication.disconnect()
            await event.call(event.Id.NEW_NOTIFICATION, 'disconnected from Lizard')



import numpy as np
from nicegui.elements.scene_object3d import Object3D
from nicegui.elements.scene_objects import Extrusion
from ..world import Area


class AreaObject(Object3D):

    def __init__(self, areas: dict[str, Area]):
        super().__init__('group')
        self.areas = areas
        self.update()

    def update(self) -> bool:
        [obj.delete() for obj in list(self.view.objects.values()) if (obj.name or '').startswith('area_')]
        for area in self.areas.values():
            if len(area.outline) == 1:
                outline = [[area.outline[0].x + 0.05 * np.cos(phi), area.outline[0].y + 0.05 * np.sin(phi)]
                           for phi in np.linspace(0, 2 * np.pi, 16, endpoint=False)]
            else:
                outline = [[point.x, point.y] for point in area.outline]
            Extrusion(outline, 0.1, wireframe=True).with_name(f'area_{area.id}').material('green')
        return False  # NOTE: avoid JustPy page_update



from collections import deque
from .. import event
from ..hardware import Hardware
from . import Actor


class Lizard(Actor):
    interval: float = 0.01

    def __init__(self, hardware: Hardware):
        super().__init__()
        self.hardware = hardware
        self.last_step = None
        event.register(event.Id.AUTOMATION_PAUSED, self._handle_pause)
        self.responsiveness_stats: deque = deque(maxlen=100)
        self.update_stats: deque = deque(maxlen=100)
        self.processing_stats: deque = deque(maxlen=100)

    async def step(self):
        await self.ensure_responsiveness()
        t = self.world.time
        await self.hardware.update()
        self.update_stats.append(self.world.time - t)
        t = self.world.time
        await event.call(event.Id.NEW_MACHINE_DATA)
        dt = self.world.time - t
        if dt > 0.02:
            self.log.warning(f'processing machine data took {dt:.2f} s')
        self.processing_stats.append(dt)

    async def _handle_pause(self, reason: str):
        await self.hardware.stop()

    async def ensure_responsiveness(self):
        dt = self.world.time - self.last_step if self.last_step is not None else 0
        if dt > 1:
            msg = f'esp serial communication can not be guaranteed ({dt:.2f} s since last step)'
            self.log.error(msg + '; aborting automation')
            await self.pause_automation(because=msg)
        elif dt > 0.1:
            self.log.warning(f'esp serial communication is slow ({dt:.2f} s since last step)')
        self.last_step = self.world.time
        self.responsiveness_stats.append(dt)



from .hardware import Hardware
from .communicating_hardware import CommunicatingHardware
from .simulated_hardware import SimulatedHardware
from .robot_brain import RobotBrain
from .robot_brain_legacy import RobotBrainLegacy



from .actor import Actor
from .asyncio_monitor import AsyncioMonitor
from .automator import Automator
from .camera_projector import CameraProjector
from .detector import Detector
from .garbage_collector import GarbageCollector
from .lizard import Lizard
from .network_monitor import NetworkMonitor, NetworkStats
from .odometer import Odometer
from .steerer import Steerer
from .usb_camera_capture import UsbCameraCapture
from .usb_camera_simulator import UsbCameraSimulator
from .backup import Backup



import numpy as np
from ..world import Velocity, World
from .hardware import Hardware


class SimulatedHardware(Hardware):

    def __init__(self, world: World):
        super().__init__(world)
        self.linear_velocity: float = 0
        self.angular_velocity: float = 0

    async def configure(self):
        raise Exception('there is no communication for configuring hardware')

    async def restart(self):
        raise Exception('there is no communication for restarting hardware')

    async def drive(self, linear: float, angular: float):
        self.linear_velocity = linear
        self.angular_velocity = angular

    async def stop(self):
        self.linear_velocity = 0
        self.angular_velocity = 0

    async def update(self):
        self.world.robot.odometry.append(Velocity(
            linear=self.linear_velocity,
            angular=self.angular_velocity,
            time=self.world.time,
        ))
        self.world.robot.battery = 25.0 + np.sin(0.1 * self.world.time) + 0.02 * np.random.randn()
        self.world.robot.temperature = np.random.uniform(34, 35)



import logging
from typing import Optional
from .. import event
from ..world import World


class Actor:
    world: World
    interval: Optional[float] = None

    def __init__(self) -> None:
        self.name = __name__[:-5] + self.__class__.__name__
        self.log = logging.getLogger(self.name)

    async def step(self):
        pass

    async def tear_down(self):
        pass

    async def pause_automation(self, *, because: str):
        await event.call(event.Id.PAUSE_AUTOMATION, because)

    async def notify(self, message: str):
        await event.call(event.Id.NEW_NOTIFICATION, message)

    def __str__(self) -> str:
        return type(self).__name__



from __future__ import annotations  # NOTE: for PEP 563 (postponed evaluation of annotations)
from pydantic import BaseModel
from typing import Any
from dataclasses import dataclass
import numpy as np
from .point import Point
from .point3d import Point3d


@dataclass
class PoseStep:
    linear: float
    angular: float
    time: float


class Pose(BaseModel):
    x: float = 0
    y: float = 0
    yaw: float = 0
    time: float = 0

    @property
    def point(self) -> Point:
        return Point(x=self.x, y=self.y)

    def point_3d(self):
        return Point3d(x=self.x, y=self.y, z=0)

    @property
    def matrix(self):
        return np.array([
            [np.cos(self.yaw), -np.sin(self.yaw), self.x],
            [np.sin(self.yaw), np.cos(self.yaw), self.y],
            [0, 0, 1],
        ])

    @property
    def yaw_deg(self):
        return np.rad2deg(self.yaw)

    @staticmethod
    def from_matrix(M, time: float = 0):
        return Pose(x=M[0, -1], y=M[1, -1], yaw=np.arctan2(M[1, 0], M[0, 0]), time=time)

    def __str__(self):
        return '%.3f, %.3f, %.1f deg' % (self.x, self.y, self.yaw_deg)

    def distance(self, other: Pose) -> float:
        return self.point.distance(other.point)

    def projected_distance(self, other: Pose) -> float:
        return self.point.projected_distance(other.point, other.yaw)

    def __iadd__(self, step: PoseStep):
        self.x += step.linear * np.cos(self.yaw)
        self.y += step.linear * np.sin(self.yaw)
        self.yaw += step.angular
        self.time = step.time
        return self

    def transform(self, point: Point) -> Point:
        return Point(
            x=self.x + point.x * np.cos(self.yaw) - point.y * np.sin(self.yaw),
            y=self.y + point.x * np.sin(self.yaw) + point.y * np.cos(self.yaw),
        )

    def transform_array(self, points: Any) -> Any:
        return np.stack((
            self.x + points[:, 0] * np.cos(self.yaw) - points[:, 1] * np.sin(self.yaw),
            self.y + points[:, 0] * np.sin(self.yaw) + points[:, 1] * np.cos(self.yaw),
        ), axis=1)

    def transform_pose(self, pose: Pose) -> Pose:
        return Pose(
            x=self.x + pose.x * np.cos(self.yaw) - pose.y * np.sin(self.yaw),
            y=self.y + pose.x * np.sin(self.yaw) + pose.y * np.cos(self.yaw),
            yaw=self.yaw + pose.yaw,
        )

    def relative_pose(self, other: Pose) -> Pose:
        return Pose.from_matrix(np.linalg.inv(self.matrix) @ other.matrix)

    def rotate(self, angle: float) -> Pose:
        return Pose(x=self.x, y=self.y, yaw=self.yaw+angle, time=self.time)



from typing import Optional
import abc
import logging


class Communication(abc.ABC):

    def __init__(self) -> None:
        self.name = __name__[:-13] + self.__class__.__name__
        self.log = logging.getLogger(self.name)
        super().__init__()

    @classmethod
    @abc.abstractmethod
    def is_possible(cls) -> bool:
        return False

    @abc.abstractmethod
    async def send_async(self, msg: str):
        pass

    @abc.abstractmethod
    async def read(self) -> Optional[str]:
        pass



from typing import Optional
import os
import serial
import logging
from .communication import Communication

log = logging.getLogger('rosys.hardware.serial_communication')


class SerialCommunication(Communication):
    baudrate: int = 115200

    def __init__(self):
        super().__init__()
        self.device_path = SerialCommunication.get_device_path()
        if self.device_path is None:
            raise Exception('No serial port found')
        self.log.debug(f'connecting serial on {self.device_path} with baudrate {self.baudrate}')
        self.serial = serial.Serial(self.device_path, self.baudrate)
        self.buffer = ''

    @classmethod
    def is_possible(cls) -> bool:
        return cls.get_device_path() is not None

    @classmethod
    def get_device_path(cls) -> Optional[str]:
        device_paths = [
            '/dev/ttyTHS1',
            '/dev/ttyUSB0',
            '/dev/tty.SLAB_USBtoUART',
            '/dev/serial/by-id/usb-Silicon_Labs_CP2102_USB_to_UART_Bridge_Controller_0001-if00-port0',
            '/dev/esp',
        ]
        for device_path in device_paths:
            if os.path.exists(device_path) and os.stat(device_path).st_gid > 0:
                return device_path

    def connect(self):
        if not self.serial.isOpen():
            self.serial.open()
            self.log.debug(f'reconnected serial on {self.device_path}')

    def disconnect(self):
        if self.serial.isOpen():
            self.serial.close()
            self.log.debug(f'disconnected serial on {self.device_path}')

    async def read(self) -> Optional[str]:
        if not self.serial.isOpen():
            return
        self.buffer += self.serial.read_all().decode()
        if '\n' in self.buffer:
            line, self.buffer = self.buffer.split('\r\n', 1)
            self.log.debug(f'read: {line}')
            return line

    async def send_async(self, line: str):
        self.serial.write(f'{line}\n'.encode())
        self.log.debug(f'send: {line}')



import logging
import abc
from ..world import World


class Hardware(abc.ABC):

    def __init__(self, world: World):
        self.world = world
        self.name = __name__[:-8] + self.__class__.__name__
        self.log = logging.getLogger(self.name)

    @abc.abstractmethod
    async def configure(self):
        '''Send current configuration to the hardware.'''
        return

    @abc.abstractmethod
    async def restart(self):
        '''Restarts hardware'''
        return

    @abc.abstractmethod
    async def drive(self, linear: float, angular: float):
        '''Send drive command to the hardware'''
        return

    @abc.abstractmethod
    async def stop(self):
        '''Stop the driving wheels'''
        return

    @abc.abstractmethod
    async def update(self):
        '''Called by actors to update the world'''
        return



import numpy as np
from rosys.world import Point
from ..world import Camera
from .actor import Actor


class CameraProjector(Actor):
    interval: float = 1.0

    async def step(self):
        for camera in self.world.cameras.values():
            if camera.calibration is None or not camera.calibration.is_complete:
                continue
            if camera.projection is not None:
                continue
            self.update_projection(camera)

    def update_projection(self, camera: Camera, rows: int = 12, columns: int = 16):
        camera.projection = [[None for _ in range(columns)] for _ in range(rows)]
        for row, j in enumerate(np.linspace(0, camera.calibration.intrinsics.size.height, rows)):
            for column, i in enumerate(np.linspace(0, camera.calibration.intrinsics.size.width, columns)):
                point = camera.calibration.project_from_image(Point(x=i, y=j))
                if point is not None:
                    camera.projection[row][column] = point.tuple[:2]



from __future__ import annotations
from typing import Optional
from uuid import uuid4
from pydantic import BaseModel, Field
import numpy as np
import abc
from .calibration import Calibration, Extrinsics, Intrinsics
from .image import Image, ImageSize
from .rotation import Rotation


class Camera(BaseModel, abc.ABC):
    id: str
    calibration: Optional[Calibration] = None
    projection: Optional[list[list[list[Optional[float]]]]] = Field(None, exclude=True)
    images: list[Image] = Field([Image.create_placeholder('no image')], exclude=True)

    @property
    def latest_image_uri(self):
        return f'camera/{self.id}/{self.images[-1].time}'

    @property
    def captured_images(self) -> list[Image]:
        return [i for i in self.images if i.data]

    @property
    def latest_captured_image(self) -> Optional[Image]:
        images = self.captured_images
        if images:
            return images[-1]

    @staticmethod
    def create_perfect_camera(
        x: float = 0, y: float = 0, z: float = 1,
        yaw: float = 0, tilt_x: float = 0, tilt_y: float = 0,
        id: str = None,
    ) -> Camera:
        calibration = Calibration(
            intrinsics=Camera.create_intrinsics(),
            extrinsics=Extrinsics(tilt=Rotation.from_euler(tilt_x, np.pi + tilt_y, 0), yaw=yaw, translation=[x, y, z]),
        )
        return Camera(
            id=id or str(uuid4()),
            calibration_simulation=calibration,
            calibration=calibration.copy(deep=True),
        )

    @staticmethod
    def create_intrinsics() -> Intrinsics:
        c = 570
        size = ImageSize(width=1600, height=1200)
        K = [[c, 0, size.width / 2], [0, c, size.height / 2], [0, 0, 1]]
        D = [0, 0, 0, 0, 0]
        rotation = Rotation.zero()
        return Intrinsics(matrix=K, distortion=D, rotation=rotation, size=size)



import asyncio
from asyncio.tasks import Task
import collections
import inspect
from aenum import Enum, auto
from typing import Awaitable, Callable, Union
import logging
import weakref
import functools


tasks: list[Task] = []
listeners = collections.defaultdict(set)
log = logging.getLogger('rosys.event')


class Id(Enum, init='value __doc__'):
    '''Event Identifier. Every event has its own set of parameters.'''

    def _generate_next_value_(name, start, count, last_values):
        '''uses enum name as value when calling auto()'''
        return name

    ROBOT_MOVED = auto(), 'triggered when a robot movement is detected'
    NEW_MACHINE_DATA = auto(), 'triggered in high frequency when machine data has been read'
    PAUSE_AUTOMATION = auto(), 'call this event to pause any running automation; provide a description of the cause as string parameter'
    AUTOMATION_PAUSED = auto(), 'triggered when the automation has been paused; provides a description of the cause as string parameter'
    NEW_NOTIFICATION = auto(), 'call this event to notify the user; provide the message as string parameter'
    NEW_CAMERA = auto(), 'called if new camera has been discoverd; provides camera as parameter'
    NEW_DETECTIONS = auto(), 'called after detection on an image is completed; provides image frame as parameter'


def register(event: Id, listener: Union[Callable, Awaitable]):
    if not callable(listener):
        raise Exception('non-callable listener')
    if listener.__name__ == '<lambda>':  # NOTE lambda functions must be stored without weakref because they will be collected otherwise
        ref = listener
    elif inspect.ismethod(listener):
        ref = weakref.WeakMethod(listener)
    else:
        ref = weakref.ref(listener)
    log.debug(f'registered {listener} to {event}')
    listeners[event].add(ref)


def unregister(event: Id, listener: Union[Callable, Awaitable]):
    marked = []
    for registered in listeners[event]:
        if hasattr(listener, '__name__') and listener.__name__ == '<lambda>' and listener == registered:
            marked.append(listener)
        if registered() == listener:
            marked.append(registered)
    [listeners[event].remove(m) for m in marked]
    log.debug(f'unregistered {listener} from {event}')


async def call(event: Id, *args):
    '''Fires event and waits async until all registered listeners are completed'''
    # if event != Id.NEW_MACHINE_DATA:
    #     log.info(f'calling {event=}')
    for listener in list(listeners.get(event, {})):
        try:
            if event != Id.NEW_MACHINE_DATA:
                log.debug(f'emitting {event} with {listener}')
            if hasattr(listener, '__name__') and listener.__name__ == '<lambda>':
                listener(*args)
                continue
            callback = listener()
            if callback is None:
                listeners[event].remove(listener)
                log.debug(f'unregistered {listener} from {event}')
                continue
            if iscoroutinefunction(listener):
                await callback(*args)
            else:
                callback(*args)
        except:
            log.exception(f'could not call {listener=} for {event=}')


def emit(event: Id, *args):
    '''Fires event without waiting for the result.'''
    assert asyncio.get_running_loop()

    # if event != Id.ROBOT_MOVED:
    #     log.info(f'emitting {event=}')
    loop = asyncio.get_event_loop()
    for listener in list(listeners.get(event, {})):
        if event != Id.ROBOT_MOVED:
            log.debug(f'emitting {event} with {listener}')
        try:
            if hasattr(listener, '__name__') and listener.__name__ == '<lambda>':
                listener(*args)
                continue
            callback = listener()
            if callback is None:
                unregister(event, listener)
                continue
            if iscoroutinefunction(listener):
                tasks.append(loop.create_task(callback(*args), name=f'handle {event.name}'))
            else:
                callback(*args)
        except:
            log.exception(f'could not call {listener=} for {event=}')


@functools.lru_cache(maxsize=100)
def iscoroutinefunction(listener):
    return inspect.iscoroutinefunction(listener())



from __future__ import annotations
from typing import List, Optional
from pydantic import BaseModel
import abc
import asyncio
from operator import ixor
from functools import reduce
from ..communication import Communication
from ..world import Velocity, World
from . import CommunicatingHardware


class RobotBrainLegacy(CommunicatingHardware):

    def __init__(self, world: World, configuration: List[HardwareGroup], communication: Communication):
        super().__init__(world, communication=communication)
        self.configuration = configuration

    async def configure(self):
        await super().configure()
        await self.send('esp erase')

        for group in self.configuration:
            for command in group.commands:
                await self.send('+' + command)
                await asyncio.sleep(0.1)

        output_modules = ','.join(group.name for group in self.configuration if group.output)
        await self.send(f'+set esp.outputModules={output_modules}')
        await self.send('+esp unmute')
        await self.send('+set esp.ready=1')
        await self.send('+set esp.24v=1')
        await self.send('esp restart')

    async def restart(self):
        await super().restart()
        await self.send('esp restart')

    async def drive(self, linear: float, angular: float):
        await super().drive(linear, angular)
        await self.send('wheels speed %.3f,%.3f' % (linear, angular))

    async def stop(self):
        await super().stop()
        await self.send('wheels power 0,0')

    async def update(self):
        await super().update()
        millis = None
        while True:
            line = await self.communication.read()
            if line is None:
                break
            if line.startswith("\x1b[0;32m"):
                self.log.warning(line)
                continue  # NOTE: ignore green log messages
            if '^' in line:
                line, checksum = line.split('^')
                if reduce(ixor, map(ord, line)) != int(checksum):
                    self.log.warning('Checksum failed')
                    continue
            if not line.startswith("esp "):
                if line.startswith('Could not send message with ID'):
                    continue
                self.log.warning(line)
                continue  # NOTE: ignore all messages but esp status
            try:
                words = line.split()[1:]
                millis = float(words.pop(0))
                if self.world.robot.clock_offset is None:
                    continue
                self.world.robot.hardware_time = millis / 1000 + self.world.robot.clock_offset
                for group in self.configuration:
                    if group.output:
                        group.parse(words, self.world)
            except (IndexError, ValueError):
                self.log.exception(f'Error parsing esp message "{line}"')

        if millis is not None:
            self.world.robot.clock_offset = self.world.time - millis / 1000

    async def send(self, msg: str):
        await self.communication.send_async(f'{msg}^{reduce(ixor, map(ord, msg))}')


class HardwareGroup(BaseModel, abc.ABC):
    name: str
    output: bool = False

    @abc.abstractproperty
    def commands(self) -> list[str]:
        pass

    def with_output(self):
        self.output = True
        return self

    def parse(self, words: list[str], world: World):
        pass


class Bluetooth(HardwareGroup):
    device_name: str

    def __init__(self, **data):
        super().__init__(name='bluetooth', **data)

    @property
    def commands(self) -> list[str]:
        return [
            f'new bluetooth {self.name} ESP_{self.device_name}',
        ]


class Can(HardwareGroup):
    rxPin: str
    txPin: str

    def __init__(self, **data):
        super().__init__(name='can', **data)

    @property
    def commands(self) -> list[str]:
        return [
            f'new can {self.name} {self.rxPin},{self.txPin}',
        ]


class Led(HardwareGroup):
    pin: str
    interval: float = 0.1
    duty: float = 0.5
    repeat: bool = True

    @property
    def commands(self) -> list[str]:
        return [
            f'new led {self.name} {self.pin}',
            f'set {self.name}.interval={self.interval}',
            f'set {self.name}.duty={self.duty}',
            f'set {self.name}.repeat={"1" if self.repeat else "0"}',
        ]


class Button(HardwareGroup):
    pin: str

    @property
    def commands(self) -> list[str]:
        return [
            f'new button {self.name} {self.pin}',
        ]


class EmergencyStop(Button):
    name: str = 'estop'

    @property
    def commands(self) -> list[str]:
        result = super().commands
        result.append(f'if {self.name} != 1 esp stop')
        return result

    def parse(self, words: list[str], world: World):
        world.robot.emergency_stop = int(words.pop(0)) == 0


class RoboClawWheels(HardwareGroup):
    address: int = 128
    baud: int = 38400
    m_per_tick: float
    width: float

    @property
    def commands(self) -> list[str]:
        return [
            f'new roboclawwheels {self.name} {self.address},{self.baud}',
            f'set {self.name}.mPerTick={self.m_per_tick}',
            f'set {self.name}.width={self.width}',
        ]

    def parse(self, words: list[str], world: World):
        world.robot.odometry.append(Velocity(
            linear=float(words.pop(0)),
            angular=float(words.pop(0)),
            time=world.robot.hardware_time,
        ))
        world.robot.temperature = float(words.pop(0))
        world.robot.battery = float(words.pop(0))


class ODriveMotor(HardwareGroup):
    can: Can
    device_id: int
    m_per_tick: float
    heartbeat_timeout: float = 2.0

    @property
    def commands(self) -> list[str]:
        return [
            f'new odrivemotor {self.name} 0,{self.can.name},{hex(self.device_id)[2:]}',
            f'set {self.name}.mPerTick={self.m_per_tick}',
            f'set {self.name}.heartbeatTimeout={self.heartbeat_timeout}',
        ]


class ODriveWheels(HardwareGroup):
    left: ODriveMotor
    right: ODriveMotor
    width: float
    left_power_factor: float = 1.0
    right_power_factor: float = 1.0

    @property
    def commands(self) -> list[str]:
        return self.left.commands + self.right.commands + [
            f'new odrivewheels {self.name} {self.left.name},{self.right.name}',
            f'set {self.name}.width={self.width}',
            f'set {self.name}.leftPowerFactor={self.left_power_factor}',
            f'set {self.name}.rightPowerFactor={self.right_power_factor}',
            # f'set {self.name}.leftTorqueFactor={self.left_power_factor}',
            # f'set {self.name}.rightTorqueFactor={self.right_power_factor}',
        ]

    def parse(self, words: list[str], world: World):
        world.robot.odometry.append(Velocity(
            linear=float(words.pop(0)),
            angular=float(words.pop(0)),
            time=world.robot.hardware_time,
        ))



from .serial_communication import SerialCommunication
from .web_communication import WebCommunication


class CommunicationFactory:

    @staticmethod
    def create():
        if SerialCommunication.is_possible():
            return SerialCommunication()
        if WebCommunication.is_possible():
            return WebCommunication()



from .. import Persistence, run
from .actor import Actor


class Backup(Actor):
    interval: float = 10

    def __init__(self, persistance: Persistence) -> None:
        super().__init__()
        self.persistance = persistance

    async def step(self):
        await super().step()
        if self.world.needs_backup:
            self.log.info('saving the world')
            await run.io_bound(self.persistance.backup)
            self.world.needs_backup = False



from .. import event
from ..world import PoseStep
from .actor import Actor


class Odometer(Actor):

    def __init__(self):
        super().__init__()

        self.last_time: float = None
        self.steps: list[PoseStep] = []
        event.register(event.Id.NEW_MACHINE_DATA, self.handle_velocity)

    def handle_velocity(self):
        if not self.world.robot.odometry:
            return
        while self.world.robot.odometry:
            velocity = self.world.robot.odometry.pop(0)

            if self.last_time is None:
                self.last_time = velocity.time
                continue

            dt = velocity.time - self.last_time
            self.last_time = velocity.time

            step = PoseStep(linear=dt*velocity.linear, angular=dt*velocity.angular, time=self.world.time)
            self.steps.append(step)
            self.world.robot.prediction += step
            self.world.robot.simulation += step

            self.world.robot.current_velocity = velocity

            if step.linear or step.angular:
                self.world.robot.last_movement = step.time
                event.emit(event.Id.ROBOT_MOVED)

        self.prune_steps(self.world.time - 10.0)

    def handle_detection(self):
        if self.world.robot.detection is None:
            return

        if self.steps and self.world.robot.detection.time < self.steps[0].time:
            return

        self.prune_steps(self.world.robot.detection.time)
        self.world.robot.prediction = self.world.robot.detection.copy(deep=True)
        for step in self.steps:
            self.world.robot.prediction += step

    def prune_steps(self, cut_off_time: float):
        while self.steps and self.steps[0].time < cut_off_time:
            del self.steps[0]



#!/usr/bin/env python3
from nicegui import ui
import pylab as pl
import time
from uuid import uuid4
from rosys.world import Obstacle, Point, Pose, Robot, World
from rosys.pathplanning.planner import Planner
import rosys.pathplanning.plot_tools as pt

world = World(robot=Robot())
for x, y, w, h in [
    [1.5, 1.5, 0.2, 6.0],
    [1.5, 1.5, 5.0, 0.2],
    [6.3, 1.5, 0.2, 2.0],
    [4.5, 2.5, 0.2, 1.5],
    [5.5, 5.0, 0.2, 0.2],
]:
    id = str(uuid4())
    outline = [Point(x=x, y=y), Point(x=x+w, y=y), Point(x=x+w, y=y+h), Point(x=x, y=y+h)]
    world.obstacles[id] = Obstacle(id=id, outline=outline)

planner = Planner(world)

plot = ui.plot()


def run():
    t = time.time()
    planner.search(goal=Pose(x=10, y=4, yaw=0))
    print('path finding: %5.1f ms' % ((time.time() - t) * 1000), flush=True)

    with plot:
        pl.clf()
        pt.show_distance_map(planner.distance_map)
        pt.show_obstacle_map(planner.obstacle_map)
        pl.gca().invert_yaxis()
        pl.autoscale(False)
        pt.plot_path(planner.raw_path, 'C0')
        [pl.plot(s.target[0], s.target[1], 'C0.') for s in planner.raw_path]
        pt.plot_path(planner.path, 'C2', lw=2)
        [pl.plot(s.target[0], s.target[1], 'C2o') for s in planner.path]
        for step in planner.path:
            pt.plot_robot(world.robot.shape, (step.spline.x(1), step.spline.y(1), step.spline.yaw(1)), 'C2', lw=1)


run()
ui.button('Run', on_click=run)

ui.run()



#!/usr/bin/env python3
from nicegui import ui
import numpy as np
import pylab as pl
from rosys.pathplanning.robot_renderer import RobotRenderer

robot_renderer = RobotRenderer.from_size(0.77, 1.21, 0.445)
pose = (0, 0, np.deg2rad(30))

with ui.scene() as scene:
    scene.extrusion(robot_renderer.outline, 0.1).move(*pose[:2]).rotate(0, 0, pose[2])

with ui.plot():
    pl.imshow(robot_renderer.render(0.1, pose[2]), cmap=pl.cm.gray, interpolation='nearest')
    pl.gca().invert_yaxis()
    pl.fill(robot_renderer.rendered_outline[:, 0],
            robot_renderer.rendered_outline[:, 1], color='none', ec='C0')

ui.run()



import asyncio
from typing import Callable, Optional, Union
from .. import Runtime, run
from ..world import Robot, World


class TestRuntime(Runtime):
    __test__ = False

    def __init__(self, world: Optional[World] = None):
        if world is None:
            world = World(robot=Robot())
        super().__init__(world)
        self.world.set_time(0)  # NOTE in tests we start at zero for better reading

        from .helper import set_global_runtime  # NOTE import here to avoid PytestAssertRewriteWarning
        set_global_runtime(self)
        self.exception = None

    async def forward(self,
                      seconds: Optional[float] = None,
                      *,
                      until: Optional[Union[int, float, Callable]] = None,
                      x: Optional[float] = None,
                      y: Optional[float] = None,
                      tolerance: float = 0.1,
                      dt: float = 0.01,
                      timeout: float = 100):
        # NOTE we start runtime here because this makes it easy in the tests to prepare it beforehand
        if not self.tasks:
            await self.startup()

        start_time = self.world.time
        robot = self.world.robot
        if seconds is not None:
            msg = f'forwarding {seconds=}'
            def condition(): return self.world.time >= start_time + seconds
            timeout = max(timeout, seconds)
        elif isinstance(until, int) or isinstance(until, float):
            msg = f'forwarding {until=}'
            def condition(): return self.world.time >= until
            timeout = max(timeout, until - start_time)
        elif isinstance(until, Callable):
            msg = f'forwarding {until=}'
            condition = until
        elif x is not None and y is not None:
            msg = f'forwarding to {x=} and {y=}'
            def condition(): return (robot.prediction.x - x)**2 + (robot.prediction.y - y)**2 < tolerance**2
        elif x is not None:
            msg = f'forwarding to {x=}'
            def condition(): return abs(robot.prediction.x - x) < tolerance
        elif y is not None:
            msg = f'forwarding to {y=}'
            def condition(): return abs(robot.prediction.y - y) < tolerance
        else:
            raise Exception('invalid arguments')

        self.log.info(f'\033[94m{msg}\033[0m')
        while not condition():
            if self.world.time > start_time + timeout:
                raise TimeoutError(f'condition not met in time')
            if not run.running_processes:
                self.world.set_time(self.world.time + dt)
                await asyncio.sleep(0)
            else:
                await asyncio.sleep(0.01)
            if self.exception is not None:
                raise RuntimeError(f'error while forwarding time {dt} s') from self.exception

    def handle_exception(self, ex: Exception):
        super().handle_exception(ex)
        self.exception = ex



from nicegui.ui import Ui

from ..hardware import CommunicatingHardware
from .. import Runtime
from .. import event
from .asyncio_page import AsyncioPage
from .automation_controls import AutomationControls
from .cpu_usage import CpuUsage
from .joystick import Joystick
from .keyboard_control import KeyboardControl
from .lizard_serial_debug import LizardSerialDebug
from .lizard_stats import LizardStats
from .objgraph_page import ObjgraphPage
from .pyloot_page import PylootPage
from .robot_object import RobotObject
from . import routes


def configure(ui: Ui, runtime: Runtime):
    ui.on_startup(runtime.startup())
    ui.on_shutdown(runtime.shutdown())
    event.register(event.Id.NEW_NOTIFICATION, ui.notify)
    routes.setup(ui, runtime)

    Joystick.steerer = runtime.steerer
    KeyboardControl.ui = ui
    KeyboardControl.steerer = runtime.steerer
    RobotObject.robot = runtime.world.robot
    RobotObject.ui = ui
    AutomationControls.runtime = runtime
    AutomationControls.ui = ui
    LizardSerialDebug.ui = ui
    if isinstance(runtime.hardware, CommunicatingHardware):
        LizardSerialDebug.communication = runtime.hardware.communication
    ObjgraphPage.ui = ui
    PylootPage.ui = ui
    CpuUsage.ui = ui
    LizardStats.ui = ui
    LizardStats.lizard = runtime.lizard
    AsyncioPage.ui = ui
    AsyncioPage.asyncio_monitor = runtime.asyncio_monitor



from nicegui.ui import Ui
import logging
import socket
import os.path
from pathlib import Path
import asyncio

log = logging.getLogger('rosys.wifi')


def has_internet() -> bool:
    '''Returns True if there's a connection'''

    IP_ADDRESS_LIST = [
        '1.1.1.1',  # Cloudflare
        '1.0.0.1',
        '8.8.8.8',  # Google DNS
        '8.8.4.4',
        '208.67.222.222',  # Open DNS
        '208.67.220.220',
    ]

    port = 53
    timeout = 3

    for host in IP_ADDRESS_LIST:
        try:
            socket.setdefaulttimeout(timeout)
            socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect((host, port))
            return True
        except socket.error:
            pass
    else:
        log.exception('No internet connection')
        return False


def create_wifi(ui: Ui):

    def add(ssid: str, password: str):
        log.info(f'adding {ssid}, {password}')
        wifi_configs = os.path.expanduser(f'~/.rosys/wifi')
        Path(wifi_configs).mkdir(parents=True, exist_ok=True)
        # NOTE a daemon on the host system must watch the .rosys/wifi dir and reload the configuration with nmcli or similar
        with open(f'{wifi_configs}/{ssid}', 'w') as f:
            f.write(password)
        dialog.close()

    async def update_wifi_status():
        if await asyncio.get_event_loop().run_in_executor(None, has_internet):
            wifi_button.props(replace='color=green')
            status.set_text(f'Robot is connected to the internet.')
        else:
            wifi_button.props(replace='color=red')
            status.set_text(f'Robot is not connected to the internet.')
        return False  # do not refresh UI

    with ui.dialog() as dialog, ui.card():
        status = ui.label()
        ssid = ui.input('ssid')
        password = ui.input('password')
        with ui.row():
            ui.button('Add', on_click=lambda: add(ssid.value, password.value))
            ui.button('Close', on_click=dialog.close)

    wifi_button = ui.button(on_click=dialog.open).props('icon=network_wifi')
    ui.timer(5, callback=update_wifi_status)



from pydantic import BaseModel
from typing import Optional
from ..helpers import ModificationContext
from .pose import Pose
from .velocity import Velocity


class RobotShape(BaseModel):
    outline: list[tuple[float, float]] = [(-0.5, -0.5), (0.5, -0.5), (0.75, 0), (0.5, 0.5), (-0.5, 0.5)]
    height: float = 0.5


class RobotParameters(BaseModel, ModificationContext):
    linear_speed_limit: float = 0.5
    angular_speed_limit: float = 0.5
    minimum_turning_radius: float = 0.0
    max_detection_age_ramp: Optional[tuple[float, float]]
    hook_offset: float = 0.5
    carrot_offset: float = 0.6
    carrot_distance: float = 0.1


class Robot(BaseModel):
    shape: RobotShape = RobotShape()
    parameters: RobotParameters = RobotParameters()
    prediction: Pose = Pose()
    detection: Optional[Pose]
    simulation: Pose = Pose()
    odometry: list[Velocity] = []
    current_velocity: Optional[Velocity]
    last_movement: float = 0
    hardware_time: Optional[float]
    battery: float = 0
    is_charging: bool = False
    temperature: float = 0
    emergency_stop: bool = False
    clock_offset: Optional[float]
    carrot: Optional[Pose]



from __future__ import annotations
from pydantic import BaseModel
import math

from .point import Point


class Point3d(BaseModel):
    x: float
    y: float
    z: float

    @property
    def tuple(self) -> tuple[float, float, float]:
        return (self.x, self.y, self.z)

    def distance(self, other: Point3d):
        return math.sqrt((other.x - self.x)**2 + (other.y - self.y)**2 + (other.z - self.z)**2)

    def projection(self) -> Point:
        return Point(x=self.x, y=self.y)

    def __str__(self) -> str:
        return f'Point3d({round(self.x, 2)}, {round(self.y, 2)}, {round(self.z, 2)})'



from nicegui.elements.scene_object3d import Object3D
from nicegui.elements.scene_objects import Extrusion
from ..world import Obstacle


class ObstacleObject(Object3D):

    def __init__(self, obstacles: dict[str, Obstacle]):
        super().__init__('group')
        self.obstacles = obstacles
        self.update()

    def update(self) -> bool:
        [obj.delete() for obj in list(self.view.objects.values()) if (obj.name or '').startswith('obstacle_')]
        for obstacle in self.obstacles.values():
            outline = [[point.x, point.y] for point in obstacle.outline]
            Extrusion(outline, 0.1).with_name(f'obstacle_{obstacle.id}')
        return False  # NOTE: avoid JustPy page_update



import logging
import json
import os
from .world import World


class Persistence:

    def __init__(self, world: World, filepath: str = '~/.rosys/world.json'):
        self.world = world
        self.filepath = os.path.expanduser(filepath)

    def dump(self) -> dict:
        return {
            'robot': {'parameters': self.world.robot.parameters.dict()},
            'usb_cameras': {id: camera.dict() for id, camera in self.world.usb_cameras.items()},
            'obstacles': {id: obstacle.dict() for id, obstacle in self.world.obstacles.items()},
            'areas': {id: area.dict() for id, area in self.world.areas.items()},
        }

    def load(self, world: World):
        self.world.robot.parameters = world.robot.parameters
        self.world.usb_cameras = world.usb_cameras
        self.world.obstacles = world.obstacles
        self.world.areas = world.areas

    def backup(self):
        os.makedirs(os.path.dirname(self.filepath), exist_ok=True)
        with open(self.filepath, 'w') as f:
            json.dump(self.dump(), f)

    def parse_world(self, obj: dict):
        return World.parse_obj(obj)

    def restore(self):
        if not os.path.exists(self.filepath):
            logging.warning('No backup file found.')
            return
        try:
            with open(self.filepath, 'r') as f:
                self.load(self.parse_world(json.load(f)))
        except:
            logging.exception(f'Could not load from backup at {self.filepath}')



from nicegui.elements.scene_objects import Extrusion, Group, Sphere, Stl
from nicegui.elements.scene_object3d import Object3D
from nicegui.ui import Ui
from ..world import Robot


class RobotObject(Object3D):
    robot: Robot = None  # will be set by rosys.ui.configure
    ui: Ui = None  # will be set by rosys.ui.configure

    def __init__(self, *, debug: bool = False):
        super().__init__('group')
        with self:
            with Group() as self.robot_group:
                outline = list(map(list, self.robot.shape.outline))
                self.robot_object = Extrusion(outline, self.robot.shape.height, wireframe=debug)
                self.robot_object.material('#4488ff', 0.5)
                if debug:
                    Sphere(0.03).material('#4488ff')
                    Sphere(0.05).material('#4488ff').move(self.robot.parameters.hook_offset)
            with Group() as self.carrot_group:
                Sphere(0.03).material('#ff8800')
                Sphere(0.05).material('#ff8800').move(self.robot.parameters.carrot_offset)
        self.ui.timer(0.05, self.update)

    def with_stl(self, url: str, *,
                 x: float = 0, y: float = 0, z: float = 0,
                 omega: float = 0, phi: float = 0, kappa: float = 0,
                 scale: float = 1.0,
                 color: str = '#ffffff', opacity: float = 1.0):
        self.robot_object.delete()
        with self.robot_group:
            self.robot_object = Stl(url).move(x, y, z).rotate(omega, phi, kappa).scale(scale).material(color, opacity)
        return self

    def update(self) -> bool:
        self.robot_group.move(self.robot.prediction.x, self.robot.prediction.y)
        self.robot_group.rotate(0, 0, self.robot.prediction.yaw)
        if self.robot.carrot is None:
            self.carrot_group.scale(0)
        else:
            self.carrot_group.scale(1)
            self.carrot_group.move(self.robot.carrot.x, self.robot.carrot.y)
            self.carrot_group.rotate(0, 0, self.robot.carrot.yaw)
        return False  # NOTE: avoid JustPy page_update



import asyncio
import inspect
import time
import sys
import numpy as np
from contextlib import contextmanager


def measure(*, reset: bool = False, ms: bool = False):
    global t
    if 't' in globals() and not reset:
        dt = time.time() - t
        line = inspect.stack()[1][0].f_lineno
        output = f'{dt * 1000:7.3f} ms' if ms else f'{dt:7.3f} s'
        print(f'{inspect.stack()[1].filename}:{line}', output, flush=True)
    if reset:
        print('------------', flush=True)
    t = time.time()


def angle(yaw0: float, yaw1: float) -> float:
    return eliminate_2pi(yaw1 - yaw0)


def eliminate_pi(angle: float) -> float:
    return (angle + np.pi / 2) % np.pi - np.pi / 2


def eliminate_2pi(angle: float) -> float:
    return (angle + np.pi) % (2 * np.pi) - np.pi


is_test: bool = 'pytest' in sys.modules


async def sleep(seconds: float):
    if is_test:
        from rosys.test.helper import global_runtime
        sleep_end_time = global_runtime.world.time + seconds
        while global_runtime.world.time <= sleep_end_time:
            await asyncio.sleep(0)
    else:
        count = int(np.ceil(seconds))
        for _ in range(count):
            await asyncio.sleep(seconds / count)


class ModificationContext:

    @contextmanager
    def set(self, **kwargs):
        backup = {key: getattr(self, key) for key in kwargs.keys()}
        for key, value in kwargs.items():
            setattr(self, key, value)
        yield
        for key, value in backup.items():
            setattr(self, key, value)



import numpy as np
import heapq
import copy
import time
import functools
from .. import run
from ..helpers import angle
from ..world import Pose, PathSegment, World
from .distance_map import DistanceMap
from .grid import Grid
from .obstacle_map import ObstacleMap
from .steps import Path, Step


class Planner:

    def __init__(self, world: World):
        self.world = world
        self.obstacle_map = None
        self.small_obstacle_map = None
        self.obstacles = None
        self.goal = None

    async def search_async(self, *, goal: Pose, start: Pose = None, backward: bool = False, timeout: float = None):
        return await run.cpu_bound(functools.partial(self.search, goal=goal, start=start, backward=backward, timeout=timeout))

    def search(self, *, goal: Pose, start: Pose = None, backward: bool = False, timeout: float = None) -> list[PathSegment]:
        if start is None:
            start = self.world.robot.prediction

        self.update_obstacle_map(goal, start)
        if self.goal != goal:
            self.distance_map = DistanceMap(self.small_obstacle_map, goal)
            self.goal = copy.deepcopy(goal)

        start_time = time.time()
        step_dist = 0.5
        num_candidates = 16

        pose = (start.x, start.y, start.yaw)
        heap = [(self.distance_map.interpolate(start.x, start.y), 0, Step(pose))]
        visited = set()

        while pose != (goal.x, goal.y, goal.yaw):
            if timeout is not None and time.time() - start_time > timeout:
                raise TimeoutError('Could not find a path')

            try:
                _, travel_cost, step = heapq.heappop(heap)
            except IndexError:
                raise TimeoutError('Could not find a path')

            pose = step.target
            tup = tuple(np.round(pose, 3))
            if tup in visited:
                continue
            visited.add(tup)

            if pose == (goal.x, goal.y, goal.yaw):
                break

            goal_dist = np.sqrt((goal.x - pose[0])**2 + (goal.y - pose[1])**2)
            goal_candidate = [(goal.x, goal.y, goal.yaw)] if goal_dist < 3 * step_dist else []
            fw_candidates = [
                (pose[0] + step_dist * np.cos(yaw), pose[1] + step_dist * np.sin(yaw), yaw)
                for yaw in np.linspace(-np.pi / 2, np.pi / 2, num_candidates // 2 + 1)[1:-1] + pose[2]
            ]
            bw_candidates = [
                (pose[0] - step_dist * np.cos(yaw), pose[1] - step_dist * np.sin(yaw), yaw)
                for yaw in np.linspace(-np.pi / 2, np.pi / 2, num_candidates // 2 + 1)[1:-1] + pose[2]
            ]

            for candidate in goal_candidate + fw_candidates + bw_candidates:
                tup = tuple(np.round(candidate, 3))
                if tup in visited:
                    continue

                if self.obstacle_map.test(*candidate):
                    continue

                costs = None
                for backward in [False, True]:
                    if backward and candidate in fw_candidates:
                        continue
                    if not backward and candidate in bw_candidates:
                        continue

                    next_step = Step(candidate, step, backward=backward)
                    if candidate in goal_candidate:
                        if not next_step.is_healthy():
                            continue

                    if self.obstacle_map.get_minimum_spline_distance(next_step.spline, backward=backward) < 0.1:
                        continue

                    if costs is None:
                        dist = self.distance_map.interpolate(candidate[0], candidate[1])[0]
                        new_travel_cost = travel_cost + step_dist
                        if step.backward != next_step.backward:
                            new_travel_cost += step_dist
                        yaw_cost = 0
                        if dist > step_dist:
                            gx, gy = self.distance_map.gradient(candidate[0], candidate[1])
                            target_yaw = np.arctan2(gy, gx) if backward else np.arctan2(-gy, -gx)
                            yaw_error = abs(angle(candidate[2], target_yaw[0]))
                            if yaw_error > np.pi / 2:
                                yaw_cost = 5 * yaw_error
                        obstacle_cost = max(1.0 - self.obstacle_map.get_distance(*candidate), 0.0)
                        remaining_cost = dist + 0.7 * new_travel_cost + yaw_cost + 0.3 * obstacle_cost
                        costs = (remaining_cost, new_travel_cost)

                    heapq.heappush(heap, (costs[0], costs[1], next_step))

        self.raw_path: Path = step.backtrace()
        self.path: Path = copy.deepcopy(self.raw_path)
        self.path.smooth(self.obstacle_map, control_dist=0.5)
        del self.path[0]
        return [PathSegment(spline=step.spline, backward=step.backward) for step in self.path]

    def update_obstacle_map(self, goal: Pose, start: Pose) -> None:
        if self.obstacles != self.world.obstacles or \
                not self.obstacle_map.grid.contains(start.point, padding=1.0) or \
                not self.obstacle_map.grid.contains(goal.point, padding=1.0):
            points = [p for obstacle in self.world.obstacles.values() for p in obstacle.outline]
            points += [p for area in self.world.areas.values() for p in area.outline]
            grid = Grid.from_points(points + [start.point, goal.point], 0.1, 36, padding=1.0)
            self.obstacle_map = ObstacleMap.from_world(self.world, grid)
            self.small_obstacle_map = self.obstacle_map  # TODO?
            self.obstacles = copy.deepcopy(self.world.obstacles)
            self.goal = None



from typing import Awaitable, Optional
from nicegui.ui import Ui
from .. import Runtime


class AutomationControls:
    ui: Ui
    runtime: Runtime

    def __init__(self, default_automation: Optional[Awaitable] = None) -> None:

        def start():
            self.runtime.automator.start(default_automation())

        def pause():
            self.runtime.automator.pause(because='pause button was pressed')

        def resume():
            self.runtime.automator.resume()

        def stop():
            self.runtime.automator.stop(because='stop button was pressed')

        play_button = self.ui.button(on_click=start).props('icon=play_arrow outline')
        pause_button = self.ui.button(on_click=pause).props('icon=pause outline')
        resume_button = self.ui.button(on_click=resume).props('icon=play_arrow outline')
        stop_button = self.ui.button(on_click=stop).props('icon=stop outline')

        def refresh():
            play_button.visible = self.runtime.automator.is_stopped
            pause_button.visible = self.runtime.automator.is_running
            resume_button.visible = self.runtime.automator.is_paused
            play_button.view.disable = default_automation is None
            stop_button.view.disable = self.runtime.automator.is_stopped
        self.ui.timer(0.1, refresh)



from pydantic import BaseModel, PrivateAttr
import time
from .. import is_test
from .area import Area
from .camera import Camera
from .obstacle import Obstacle
from .robot import Robot
from .upload import Upload
from .usb_camera import UsbCamera


class World(BaseModel):
    robot: Robot = Robot()
    _time: float = PrivateAttr(default_factory=time.time)
    obstacles: dict[str, Obstacle] = {}
    areas: dict[str, Area] = {}
    notifications: list[tuple[float, str]] = []
    usb_cameras: dict[str, UsbCamera] = {}
    upload: Upload = Upload()
    needs_backup: bool = False

    @property
    def time(self) -> float:
        return self._time if is_test else time.time()

    def set_time(self, value):
        assert is_test
        self._time = value

    @property
    def cameras(self) -> dict[str, Camera]:
        return self.usb_cameras



from typing import Any, Union
import pytest
import numpy as np
from ..automations import drive_to
from ..world import Point, Point3d
from .runtime import TestRuntime

global_runtime: TestRuntime = None


def set_global_runtime(runtime: TestRuntime):
    global global_runtime
    global camera_count
    global_runtime = runtime
    camera_count = 0


def assert_pose(x: float, y: float, *, deg: float = None, linear_tolerance: float = 0.1, deg_tolerance: float = 1.0):
    pose = global_runtime.world.robot.prediction
    assert pose.x == pytest.approx(x, abs=linear_tolerance)
    assert pose.y == pytest.approx(y, abs=linear_tolerance)

    if deg is not None:
        assert np.rad2deg(pose.yaw) == pytest.approx(deg, abs=deg_tolerance)


def assert_point(actual: Union[Point, Point3d], expected: Union[Point, Point3d], tolerance=0.1):
    assert type(actual) == type(expected)

    assert actual.x == pytest.approx(expected.x, abs=tolerance)
    assert actual.y == pytest.approx(expected.y, abs=tolerance)

    if type(actual) is Point3d:
        assert actual.z == pytest.approx(expected.z, abs=tolerance)


async def automate_drive_to(x: float, y: float):
    global_runtime.automator.start(drive_to(global_runtime.world, global_runtime.hardware, Point(x=x, y=y)))


def approx(o1: Any, o2: Any) -> None:
    # code is taken from https://github.com/pytest-dev/pytest/issues/6632#issuecomment-580507745
    assert type(o1) == type(o2)

    o1_keys = [v for v in dir(o1) if not v.startswith('__')]
    o2_keys = [v for v in dir(o2) if not v.startswith('__')]

    assert sorted(o1_keys) == sorted(o2_keys)

    for k in o1_keys:
        v1 = getattr(o1, k)
        v2 = getattr(o2, k)
        if isinstance(v1, int) or isinstance(v1, float):
            assert v1 == pytest.approx(v2)
            continue

        if isinstance(v1, bool) or isinstance(v1, str):
            assert v1 == v2
            continue

        approx(v1, v2)



from nicegui.ui import Ui
import starlette
from rosys import Runtime
import logging

log = logging.getLogger('rosys.routes')
not_found = starlette.responses.Response(content='Not Found', status_code=404)


def setup(ui: Ui, runtime: Runtime):
    ui.add_route(starlette.routing.Route(
        '/world', lambda request, **_:
        starlette.responses.Response(content=runtime.world.json(), media_type='text/json')))
    ui.add_route(starlette.routing.Route(
        '/export', lambda request, **_:
        starlette.responses.JSONResponse(content=runtime.persistence.dump())))

    def get_image(request, **_):
        try:
            cam_id = request.path_params['id']
            if cam_id not in runtime.world.usb_cameras:
                return not_found
            for image in reversed(runtime.world.usb_cameras[cam_id].images):
                if str(image.time) == request.path_params['timestamp']:
                    return starlette.responses.Response(content=image.data, media_type='image/jpeg')
            return not_found
        except:
            log.exception('could not get image')
            raise

    ui.add_route(starlette.routing.Route('/camera/{id}/{timestamp}', get_image))



from .configure import configure

from .area_object import AreaObject as area_object
from .asyncio_page import AsyncioPage as asyncio_page
from .automation_controls import AutomationControls as automation_controls
from .cpu_usage import CpuUsage as cpu_usage
from .joystick import Joystick as joystick
from .keyboard_control import KeyboardControl as keyboard_control
from .lizard_serial_debug import LizardSerialDebug as lizard_serial_debug
from .lizard_stats import LizardStats as lizard_stats
from .objgraph_page import ObjgraphPage as objgraph_page
from .obstacle_object import ObstacleObject as obstacle_object
from .path_object import PathObject as path_object
from .profiler import create_profiler
from .pyloot_page import PylootPage as pyloot_page
from .robot_object import RobotObject as robot_object
from .wifi import create_wifi



import cv2
import re
import shutil
import rosys
from ..world import UsbCamera, Image, ImageSize
from .. import event
from .actor import Actor


class UsbCameraCapture(Actor):
    interval: float = 0.30

    def __init__(self):
        super().__init__()
        self.devices = {}  # mapping camera ids to opencv devices
        self.last_scan = None

    async def step(self):
        await super().step()
        await self.update_device_list()
        for uid, camera in self.world.usb_cameras.items():
            if not camera.capture or not camera.connected:
                return
            bytes = await rosys.run.io_bound(self.capture_image, uid)
            camera.images.append(Image(camera_id=uid, data=bytes, time=self.world.time, size=camera.resolution))
        self.purge_old_images()

    def capture_image(self, id) -> bytes:
        _, image = self.devices[id].read()
        ic(image)
        return cv2.imencode('.jpg', image)[1].tobytes()

    def purge_old_images(self):
        for camera in self.world.usb_cameras.values():
            while camera.images and camera.images[0].time < self.world.time - 5 * 60.0:
                del camera.images[0]

    async def update_device_list(self):
        if self.last_scan is not None and self.world.time < self.last_scan + 30:  # scan every 30 sec
            return
        self.last_scan = self.world.time
        output = await rosys.run.sh(['v4l2-ctl', '--list-devices'])
        for c in self.world.usb_cameras.values():
            c.connected = False
        for infos in output.split('\n\n'):
            if 'Cannot open device' in infos:
                continue
            match = re.search('\((.*)\)', infos)
            if match is None:
                continue
            uid = match.group(1)
            if uid not in self.world.usb_cameras:
                self.world.usb_cameras[uid] = UsbCamera(id=uid, connected=True)
                self.log.info(f'adding camera {uid}')
                await event.call(event.Id.NEW_CAMERA, self.world.usb_cameras[uid])
            self.world.usb_cameras[uid].connected = True
            lines = infos.splitlines()
            num = int(lines[1].strip().lstrip('/dev/video'))
            if uid not in self.devices:
                device = self.get_capture_device(num)
                if device is None:
                    if uid in self.world.usb_cameras:
                        self.world.usb_cameras[uid].connected = False
                else:
                    self.devices[uid] = device
            if uid in self.devices:
                await self.update_parameters(uid, num)

    def get_capture_device(self, index: int):
        try:
            capture = cv2.VideoCapture(index)
            if capture is None or not capture.isOpened():
                self.log.error(f'{index} is unavailable device')
            else:
                return capture
        except:
            self.log.exception(f'{index} device failed')

    async def tear_down(self):
        await super().tear_down()
        for capture in self.devices.values():
            capture.release()

    @staticmethod
    def is_operable() -> bool:
        return shutil.which('v4l2-ctl') is not None

    async def update_parameters(self, uid: str, dev_num: int):
        output = await rosys.run.sh(['v4l2-ctl', '--all', '-d', str(dev_num)])
        size = re.search('Width/Height.*: (\d*)/(\d*)', output)
        self.world.usb_cameras[uid].resolution = ImageSize(width=int(size.group(1)), height=int(size.group(2)))



import socketio
from datetime import datetime, timedelta
from .. import event, sleep, task_logger
from ..world import BoxDetection, Image, PointDetection
from .actor import Actor


class Detector(Actor):
    interval: float = 0

    def __init__(self):
        super().__init__()
        self.sio = socketio.AsyncClient()
        self.is_connected = None

        @self.sio.event
        def connect():
            self.is_connected = True
            assert self.sio.transport() == 'websocket'

        @self.sio.event
        def connect_error(data):
            self.is_connected = False

        @self.sio.event
        def disconnect():
            self.is_connected = False

    async def connect(self, reconnect_delay: float = 3):
        try:
            self.log.info('connecting to detector')
            await self.sio.connect('ws://localhost:8004', socketio_path='/ws/socket.io')
            self.log.info('connected successfully')
            self.is_connected = True
        except:
            self.log.exception('connection failed; trying again')
            self.is_connected = None
            await sleep(reconnect_delay)

    async def step(self):
        if self.is_connected is None:
            await self.connect()

        if not self.is_connected:
            return

        for image in self.world.upload.queue:
            if datetime.now() < self.world.upload.last_upload + timedelta(minutes=self.world.upload.minimal_minutes_between_uploads):
                break
            if image.data:
                self.world.upload.last_upload = datetime.now()
                self.world.upload.queue.clear()  # because old images should not be uploaded later when the robot is inactive
                task_logger.create_task(self.upload(image), name='upload_image')
                break

        detecting_cameras = [c for c in self.world.usb_cameras.values() if c.detect]
        if not detecting_cameras:
            await sleep(0.02)
            return

        for camera in detecting_cameras:
            image = camera.images[-1]
            await self.detect(image)

    async def detect(self, image: Image) -> Image:
        if not self.is_connected:
            return
        try:
            result = await self.sio.call('detect', {'image': image.data, 'mac': image.camera_id})
            box_detections = [BoxDetection.parse_obj(d) for d in result.get('box_detections', [])]
            point_detections = [PointDetection.parse_obj(d) for d in result.get('point_detections', [])]
            image.detections = box_detections + point_detections
        except:
            self.log.exception(f'could not detect {image.id}')
        else:
            event.emit(event.Id.NEW_DETECTIONS, image)
            return image

    async def upload(self, image: Image):
        try:
            await self.sio.emit('upload', {'image': image.data, 'mac': image.camera_id})
        except:
            self.log.exception(f'could not upload {image.id}')

    def __str__(self) -> str:
        state = {
            None: 'starting',
            True: 'connected',
            False: 'reconnecting',
        }[self.is_connected]
        return f'{type(self).__name__} ({state})'



from enum import Enum
from .. import event
from ..hardware import Hardware
from . import Actor


class State(Enum):
    IDLE = 1
    INITIALIZING = 2
    STEERING = 3
    STOPPING = 4


class Steerer(Actor):
    interval: float = 0.05

    def __init__(self, hardware: Hardware):
        super().__init__()
        self.hardware = hardware
        self.state = State.IDLE
        self.orientation = None
        self.linear_speed = 0
        self.angular_speed = 0

    def start(self):
        self.log.info('start steering')
        self.state = State.INITIALIZING
        event.emit(event.Id.PAUSE_AUTOMATION, 'using steerer')

    def update(self, x: float, y: float):
        if self.state == State.INITIALIZING:
            squared_distance = x**2 + y**2
            dead_zone = 0.1
            if squared_distance > dead_zone**2:
                is_down = y < 0 and abs(y) > abs(x)
                self.orientation = -1 if is_down else 1
                self.state = State.STEERING

        if self.state == State.STEERING:
            self.linear_speed = y
            self.angular_speed = -x * self.orientation

    def stop(self):
        self.log.info('stop steering')
        self.orientation = None
        self.state = State.STOPPING

    async def step(self):
        if self.state == State.STEERING:
            await self.hardware.drive(self.linear_speed, self.angular_speed)
        elif self.state == State.STOPPING:
            await self.hardware.drive(0, 0)
            self.state = State.IDLE

    def __str__(self) -> str:
        return f'{type(self).__name__} ({self.state})'



from .helpers import is_test, sleep
from .persistence import Persistence
from .runtime import Runtime
from . import actors
from . import communication
from . import event
from . import hardware
from . import pathplanning
from . import run
from . import world



from nicegui.ui import Ui
from nicegui.elements.chart import Chart
import psutil


class CpuUsage(Chart):
    ui: Ui  # will be set by rosys.ui.configure

    def __init__(self) -> None:
        super().__init__({
            'title': {'text': 'CPU', 'floating': True, 'y': 20},
            'chart': {'type': 'line', 'animation': False},
            'xAxis': {'labels': False},
            'yAxis': {'min': 0, 'max': 100, 'title': {'text': '%'}},
            'series': [{'data': [p]} for p in psutil.cpu_percent(percpu=True)],
            'plotOptions': {'series': {'marker': False}},
            'navigation': {'buttonOptions': {'enabled': False}},
            'legend': False,
            'credits': False,
        })
        self.ui.timer(0.1, self.update)

    def update(self):
        for i, v in enumerate(psutil.cpu_percent(percpu=True)):
            self.options.series[i].data.append(v)
        [s.data.pop(0) for s in self.options.series if len(s.data) > 20]



from ..hardware import Hardware
from ..world import Point, World
from . import drive_to


async def drive_square(world: World, hardware: Hardware):
    start_pose = world.robot.prediction.copy()
    for x, y in [(1, 0), (1, 1), (0, 1), (0, 0)]:
        await drive_to(world, hardware, start_pose.transform(Point(x=x, y=y)))



from nicegui.elements.joystick import Joystick as NiceGuiJoystick
from ..actors import Steerer


class Joystick(NiceGuiJoystick):
    steerer: Steerer = None  # will be set by rosys.ui.configure

    def __init__(self, **options):
        super().__init__(on_start=self.handle_start, on_move=self.handle_move, on_end=self.handle_end, **options)

    def handle_start(self, _):
        self.steerer.start()
        return False

    def handle_move(self, msg):
        self.steerer.update(msg.data.vector.x, msg.data.vector.y)
        return False

    def handle_end(self, _):
        self.steerer.stop()
        return False



from pydantic import BaseModel


class Detection(BaseModel):
    category_name: str
    model_name: str
    confidence: float
    x: float
    y: float

    @property
    def cx(self) -> float:
        raise NotImplementedError()

    @property
    def cy(self) -> float:
        raise NotImplementedError()


class BoxDetection(Detection):

    width: float
    height: float

    @property
    def cx(self) -> float:
        return self.x + self.width / 2

    @property
    def cy(self) -> float:
        return self.y + self.height / 2

    def __str__(self):
        return f'x:{self.x:.0f} y: {self.y:.0f}, w: {self.width:.0f} h: {self.height:.0f} cat: {self.category_name} conf: {self.confidence:.2f}'


class PointDetection(Detection):

    @property
    def cx(self) -> float:
        return self.x

    @property
    def cy(self) -> float:
        return self.y

    def __str__(self):
        return f'x:{self.x:.0f} y: {self.y:.0f} cat: {self.category_name} conf: {self.confidence:.2f}'



from pydantic import BaseModel, Field
from datetime import datetime
from .image import Image


class Upload(BaseModel):
    minimal_minutes_between_uploads: float = 1.0
    last_upload: datetime = datetime.fromtimestamp(0)
    queue: set[Image] = Field(set(), exclude=True)

    def mark(self, image: Image):
        '''Mark an image for upload. May still not get uploaded because minimal_minutes_between_uploads have not elapsed.'''
        self.queue.add(image)



from nicegui.ui import Ui
from nicegui.elements.image import Image
import objgraph
import random
import base64
import asyncio
import logging
import gc
from collections import Counter

log = logging.getLogger('rosys.objgraph_page')


class ObjgraphPage:
    ui: Ui = None  # will be set by rosys.ui.configure

    def __init__(self) -> None:
        with self.ui.page('/objgraph'):
            t = self.ui.timer(10, self.refresh_stats, active=False)
            self.ui.switch('track objects (every 10 s)').bind_value_to(t, 'active')
            with self.ui.column():
                self.growth = self.ui.label()
                self.leaking = self.ui.label()
                self.overall = self.ui.label()
                self.counts = self.ui.markdown()

            self.class_search = self.ui.input('Search Class', on_change=self.update_graph)
            self.class_search_result = self.ui.image().props('contain').style('height:400px;width:100%')
            self.class_search_result.visible = False
            self.most_common_search_result = self.ui.image().props('contain').style('height:400px;width:100%')
            self.most_common_search_result.visible = False

    def update_graph(self, _=None):
        if not self.class_search.value:
            return
        objgraph.show_chain(objgraph.find_backref_chain(
            random.choice(objgraph.by_type(self.class_search.value)),  # just pick one of the found objects
            objgraph.is_proper_module),
            filename='/tmp/rosys_objgraph_class_search.png')
        self.load_image('/tmp/rosys_objgraph_class_search.png', self.class_search_result)

    def load_image(self, file: str, img: Image):
        try:
            with open(file, 'rb') as f:
                data = base64.b64encode(f.read()).decode('utf-8')
            img.clear()
            img.visible = True
            img.set_source('data:image/png;base64,'+str(data))
        except:
            log.exception('could not load image ' + file)

    def get_objgraph_stats(self):
        gc.collect()
        growth_objs = objgraph.growth(4)
        growth = 'object growth every 10 sec: '
        for obj in growth_objs:
            growth += f'{obj[0]}: +{obj[2]}  '
        leaking = 'obj without refs: '
        for o, c in list(objgraph.typestats(objgraph.get_leaking_objects()).items())[:4]:
            leaking += f'{o}: {c}  '
        overall = 'overall obj count: '
        for o, c in list(objgraph.typestats().items())[:4]:
            overall += f'{o}: {c}  '
        if growth_objs:
            biggest_growth = growth_objs[0][0]
            new = objgraph.at_addrs(objgraph.get_new_ids()[biggest_growth])
            counts = Counter(str(i) for i in new)
            if counts:
                most_common = counts.most_common()[0][0]
                unique = {str(i): i for i in new}
                # build chain for most common object
                objgraph.show_chain(objgraph.find_backref_chain(
                    unique[most_common],
                    objgraph.is_proper_module),
                    filename='/tmp/rosys_objgraph_most_common.png'
                )

        return growth, leaking, overall, counts

    async def refresh_stats(self):
        loop = asyncio.get_running_loop()
        growth, leaking, overall, counts = await loop.run_in_executor(None, self.get_objgraph_stats)
        await loop.run_in_executor(None, self.update_graph)
        self.growth.set_text(growth)
        self.leaking.set_text(leaking)
        self.overall.set_text(overall)
        highest = [f'{c[1]}: {c[0]}' for c in counts.most_common()[:5]]
        self.counts.set_content('new obj occurrence: ' + '\n'.join(highest))
        if counts:
            self.load_image('/tmp/rosys_objgraph_most_common.png', self.most_common_search_result)
        else:
            self.most_common_search_result.visible = False



from ..world import World
from ..communication import Communication
from .hardware import Hardware


class CommunicatingHardware(Hardware):

    def __init__(self, world: World, communication: Communication):
        super().__init__(world)
        self.communication = communication



from __future__ import annotations
from asyncio.exceptions import CancelledError
import asyncio
import logging
from typing import Optional, Type

from . import event, task_logger, run, sleep, is_test, Persistence
from .actors import Actor, Automator, Lizard, Odometer, Steerer, UsbCameraCapture, UsbCameraSimulator, NetworkMonitor, Backup, AsyncioMonitor, GarbageCollector
from .hardware import Hardware, SimulatedHardware
from .world import World


class Runtime:
    log = logging.getLogger(__name__)

    def __init__(self,
                 world: Optional[World] = None,
                 persistence: Optional[Persistence] = None,
                 hardware: Optional[Hardware] = None):
        self.world = world or World()
        Actor.world = self.world
        self.tasks = []
        if not is_test:
            self.persistence = persistence or Persistence(self.world)
            self.persistence.restore()
        self.hardware = hardware or SimulatedHardware(self.world)
        self.lizard = Lizard(self.hardware)
        self.odometer = Odometer()
        self.steerer = Steerer(self.hardware)
        self.automator = Automator()
        self.asyncio_monitor = AsyncioMonitor()
        self.actors = [
            self.lizard,
            self.odometer,
            self.steerer,
            self.automator,
            self.asyncio_monitor,
            GarbageCollector(),
        ]
        if NetworkMonitor.is_operable():
            self.with_actors(NetworkMonitor())
        if not is_test:
            self.with_actors(Backup(self.persistence))

    def with_actors(self, *actors: list[Actor]) -> Runtime:
        '''Adds list of additional actors to runtime.'''
        self.actors += actors
        return self

    def with_usb_cameras(self) -> Runtime:
        '''Adds usb camera capture actor to runtime.'''
        if UsbCameraCapture.is_operable() and not is_test:
            self.with_actors(UsbCameraCapture())
        else:
            self.with_actors(UsbCameraSimulator())
        return self

    async def startup(self):
        if self.tasks:
            raise Exception('should be only executed once')

        event.register(event.Id.NEW_NOTIFICATION, self.store_notification)
        for actor in self.actors:
            if actor.interval is not None:
                self.log.debug(f'starting actor {actor.name} with interval {actor.interval}s')
                self.tasks.append(task_logger.create_task(self.repeat(actor), name=actor.name))
        self.tasks.append(asyncio.create_task(self.watch_emitted_events()))

        if not is_test:
            await asyncio.sleep(1)  # NOTE we wait for RoSys to start up before analyzing async debugging
        self.activate_async_debugging()
        self.log.debug('startup completed')

    async def shutdown(self):
        await self.hardware.drive(0, 0)
        if not is_test:
            self.persistence.backup()
        [t.cancel() for t in self.tasks]
        if not is_test:
            run.process_pool.shutdown()
        await asyncio.gather(*[task_logger.create_task(a.tear_down(), name=f'{a.name}.tear_down()') for a in self.actors])

    async def repeat(self, actor: Actor):
        while True:
            start = self.world.time
            try:
                await actor.step()
                dt = self.world.time - start
            except (CancelledError, GeneratorExit):
                return
            except:
                dt = self.world.time - start
                self.log.exception(f'error in {actor}')
                if actor.interval == 0 and dt < 0.1:
                    delay = 0.1 - dt
                    self.log.warning(
                        f'{actor} would be called to frequently ' +
                        f'because it only took {dt*1000:.0f} ms; ' +
                        f'delaying this step for {delay*1000:.0f} ms')
                    await sleep(delay)
            try:
                await sleep(actor.interval - dt)
            except (CancelledError, GeneratorExit):
                return

    def get_actor(self, _type: Type):
        return next((a for a in self.actors if type(a) is _type), None)

    def store_notification(self, message: str):
        self.log.info(message)
        self.world.notifications.append((self.world.time, message))

    def activate_async_debugging(self):
        '''Produce warnings for coroutines which take too long on the main loop and hence clog the event loop'''
        try:
            loop = asyncio.get_running_loop()
            loop.set_debug(True)
            loop.slow_callback_duration = 0.05
        except:
            self.log.exception('could not activate async debugging')

    async def watch_emitted_events(self):
        while True:
            try:
                for task in event.tasks:
                    if task.done() and task.exception():
                        self.handle_exception(task.exception())
                # cleanup finished tasks
                event.tasks = [t for t in event.tasks if not t.done()]
            except:
                self.log.exception('failing to watch emitted events')
            await sleep(0 if is_test else 0.1)

    def handle_exception(self, ex: Exception):
        self.log.exception('task failed to execute', exc_info=ex)



from .area import Area
from .calibration import Calibration, Intrinsics, Extrinsics
from .camera import Camera
from .detection import Detection, BoxDetection, PointDetection
from .image import Image, ImageSize
from .line import Line
from .line_segment import LineSegment
from .obstacle import Obstacle
from .path_segment import PathSegment
from .point import Point
from .point3d import Point3d
from .pose import Pose, PoseStep
from .robot import Robot, RobotParameters, RobotShape
from .rotation import Rotation
from .spline import Spline
from .upload import Upload
from .usb_camera import UsbCamera
from .velocity import Velocity
from .world import World
